@@ -49,9 +49,13 @@
 
 package com.openexchange.pns.impl;
 
+import static com.openexchange.java.Autoboxing.Coll2i;
 import static com.openexchange.java.Autoboxing.I;
+import static com.openexchange.java.Autoboxing.L;
+import static com.openexchange.logging.LogUtility.toStringObjectFor;
 import java.util.Collection;
 import java.util.Collections;
+import java.util.HashMap;
 import java.util.Iterator;
 import java.util.LinkedHashMap;
 import java.util.LinkedList;
@@ -70,6 +74,7 @@ import com.openexchange.pns.PushMatch;
 import com.openexchange.pns.PushNotification;
 import com.openexchange.pns.PushNotificationService;
 import com.openexchange.pns.PushNotificationTransport;
+import com.openexchange.pns.PushPriority;
 import com.openexchange.pns.PushSubscriptionRegistry;
 import com.openexchange.processing.Processor;
 import com.openexchange.processing.ProcessorService;
@@ -108,6 +113,36 @@ public class PushNotificationServiceImpl implements PushNotificationService {
         return tmp.longValue();
     }
 
+    private static long delayDuration(ConfigurationService configService, PushPriority priority) {
+        long baseDelayDuration = delayDuration(configService);
+        return null != priority ? priority.getDelay(baseDelayDuration) : baseDelayDuration;
+    }
+
+    private static volatile Long maxDelayDuration;
+
+    private static long maxDelayDuration(ConfigurationService configService) {
+        Long tmp = maxDelayDuration;
+        if (null == tmp) {
+            synchronized (PushNotificationServiceImpl.class) {
+                tmp = maxDelayDuration;
+                if (null == tmp) {
+                    int defaultValue = 10000; // 10 seconds
+                    if (null == configService) {
+                        return defaultValue;
+                    }
+                    tmp = Long.valueOf(configService.getIntProperty("com.openexchange.pns.maxDelayDuration", defaultValue));
+                    maxDelayDuration = tmp;
+                }
+            }
+        }
+        return tmp.longValue();
+    }
+
+    private static long maxDelayDuration(ConfigurationService configService, PushPriority priority) {
+        long baseMaxDelayDuration = maxDelayDuration(configService);
+        return null != priority ? priority.getDelay(baseMaxDelayDuration) : baseMaxDelayDuration;
+    }
+
     private static volatile Long timerFrequency;
 
     private static long timerFrequency(ConfigurationService configService) {
@@ -188,7 +223,7 @@ public class PushNotificationServiceImpl implements PushNotificationService {
 
     private final Processor processor;
 
-    private final UnsynchronizedBufferingQueue<PushNotification> scheduledNotifcations; // Accessed synchronized
+    private final UnsynchronizedBufferingQueue<PushNotification> submittedNotifcations; // Accessed synchronized
     private ScheduledTimerTask scheduledTimerTask; // Accessed synchronized
     private boolean stopped; // Accessed synchronized
 
@@ -211,7 +246,7 @@ public class PushNotificationServiceImpl implements PushNotificationService {
         this.configService = configService;
         this.timerService = timerService;
         this.transportRegistry = transportRegistry;
-        scheduledNotifcations = new UnsynchronizedBufferingQueue<>(delayDuration(configService));
+        submittedNotifcations = new UnsynchronizedBufferingQueue<>(delayDuration(configService), maxDelayDuration(configService));
         numOfProcessedNotifications = new AtomicLong(0);
         numOfSubmittedNotifications = new AtomicLong(0);
     }
@@ -225,7 +260,7 @@ public class PushNotificationServiceImpl implements PushNotificationService {
                 return 0L;
             }
 
-            return scheduledNotifcations.size();
+            return submittedNotifcations.size();
         } finally {
             lock.unlock();
         }
@@ -255,13 +290,12 @@ public class PushNotificationServiceImpl implements PushNotificationService {
             }
 
             if (handleRemaining) {
-                for (PushNotification notification : scheduledNotifcations) {
-                    int userId = notification.getUserId();
-                    int contextId = notification.getContextId();
+                for (PushNotification notification : submittedNotifcations) {
                     try {
-                        doHandle(Collections.singleton(notification), notification.getTopic(), 1, userId, contextId);
+                        new NotificationsHandler(notification).run();
                     } catch (Exception e) {
-                        LOGGER.error("Failed to handle notification with topic {} for user {} in context {}", notification.getTopic(), I(userId), I(contextId), e);
+                        LOGGER.error("Failed to handle notification with topic {} for user {} in context {}",
+                            notification.getTopic(), I(notification.getUserId()), I(notification.getContextId()), e);
                     }
                 }
 
@@ -275,7 +309,7 @@ public class PushNotificationServiceImpl implements PushNotificationService {
                 processor.stop();
             }
 
-            scheduledNotifcations.clear();
+            submittedNotifcations.clear();
             cancelTimerTask();
             stopped = true;
         } finally {
@@ -289,50 +323,25 @@ public class PushNotificationServiceImpl implements PushNotificationService {
      * May only be accessed when holding lock.
      */
     protected void cancelTimerTask() {
+        ScheduledTimerTask scheduledTimerTask = this.scheduledTimerTask;
         if (null != scheduledTimerTask) {
+            this.scheduledTimerTask = null;
             scheduledTimerTask.cancel();
-            scheduledTimerTask = null;
         }
     }
 
     @Override
-    public void handle(PushNotification notification) throws OXException {
-        lock.lock();
-        try {
-            if (stopped) {
-                return;
-            }
-
-            // Add to queue
-            addToQueue(notification);
-
-            // Fire off worker if paused
-            if (null == scheduledTimerTask) {
-                Runnable task = new Runnable() {
-
-                    @Override
-                    public void run() {
-                        checkNotifications();
-                    }
-                };
-                long initialDelay = delayDuration(configService);
-                long delay = timerFrequency(configService);
-                scheduledTimerTask = timerService.scheduleWithFixedDelay(task, initialDelay, delay);
-            }
-        } finally {
-            lock.unlock();
-        }
-    }
-
-    @Override
-    public void handle(Collection<PushNotification> notifications) throws OXException {
+    public void handle(Collection<PushNotification> notifications, PushPriority priority) throws OXException {
         if (null == notifications || notifications.isEmpty()) {
             return;
         }
 
+        long delayDuration = delayDuration(configService, priority); // Default 1 second
+        long maxDelayDuration = maxDelayDuration(configService, priority); // Default 10 seconds
         lock.lock();
         try {
             if (stopped) {
+                LOGGER.debug("Push notification service stopped. Aborting enqueueing of notifications...");
                 return;
             }
 
@@ -341,7 +350,7 @@ public class PushNotificationServiceImpl implements PushNotificationService {
 
             // Grab first one for bootstrapping
             PushNotification firstNotification = iterator.next();
-            addToQueue(firstNotification);
+            addToQueue(firstNotification, delayDuration, maxDelayDuration);
 
             // Fire off worker if paused
             if (null == scheduledTimerTask) {
@@ -352,29 +361,30 @@ public class PushNotificationServiceImpl implements PushNotificationService {
                         checkNotifications();
                     }
                 };
-                long initialDelay = delayDuration(configService);
-                long delay = timerFrequency(configService);
+                long delay = timerFrequency(configService); // Default 0,5 second
+                long initialDelay = delayDuration + 500L; // Default 1 second + 0,5 second
                 scheduledTimerTask = timerService.scheduleWithFixedDelay(task, initialDelay, delay);
+                LOGGER.debug("Initiated polling timer task with initial delay of {}msec and periodic delay of {}msec", L(initialDelay), L(delay));
             }
 
             // Add rest to queue
             while (iterator.hasNext()) {
-                addToQueue(iterator.next());
+                addToQueue(iterator.next(), delayDuration, maxDelayDuration);
             }
         } finally {
             lock.unlock();
         }
     }
 
-    private void addToQueue(PushNotification notification) {
-        boolean added = scheduledNotifcations.offerIfAbsentElseReset(notification);
+    private void addToQueue(PushNotification notification, long delayDuration, long maxDelayDuration) {
+        boolean added = submittedNotifcations.offerIfAbsentElseReset(notification, delayDuration, maxDelayDuration);
         if (added) {
             if (numOfSubmittedNotifications.incrementAndGet() < 0L) {
                 numOfSubmittedNotifications.set(0L);
             }
-            LOGGER.debug("Scheduled notification \"{}\" for user {} in context {}", notification.getTopic(), I(notification.getUserId()), I(notification.getContextId()));
+            LOGGER.debug("Enqueued notification \"{}\" for user {} in context {} with delay of {}msec ({}msec max. delay)", notification.getTopic(), I(notification.getUserId()), I(notification.getContextId()), L(delayDuration), L(maxDelayDuration));
         } else {
-            LOGGER.debug("Reset & re-scheduled notification \"{}\" for user {} in context {}", notification.getTopic(), I(notification.getUserId()), I(notification.getContextId()));
+            LOGGER.debug("Reset & re-enqueued notification \"{}\" for user {} in context {} with delay of {}msec ({}msec max. delay)", notification.getTopic(), I(notification.getUserId()), I(notification.getContextId()), L(delayDuration), L(maxDelayDuration));
         }
     }
 
@@ -382,58 +392,82 @@ public class PushNotificationServiceImpl implements PushNotificationService {
      * Checks for an available notifications.
      */
     protected void checkNotifications() {
-        List<PushNotification> polled = null;
-        int numAdded = 0;
+        List<PushNotification> polled;
+        boolean onlyOne;
 
         // Leave this mutually exclusive section as fast as possible...
         lock.lock();
         try {
             if (stopped) {
+                LOGGER.debug("Push notification service stopped. Aborting polling of queue...");
                 return;
             }
 
-            PushNotification notification = scheduledNotifcations.poll();
-            if (null == notification) {
+            PushNotification first = submittedNotifcations.poll();
+            if (null == first) {
                 // Queue has no notification with an expired delay
+                LOGGER.debug("Queue has no notification with an expired delay and holds {} buffered notifications", I(submittedNotifcations.size()));
                 return;
             }
 
             // Check for more
-            polled = new LinkedList<>();
-            do {
-                polled.add(notification);
-                numAdded++;
-                notification = scheduledNotifcations.poll();
-            } while (null != notification);
+            PushNotification notification = submittedNotifcations.poll();
+            if (notification != null) {
+                polled = new LinkedList<>();
+                polled.add(first);
+                LOGGER.debug("Polled notification \"{}\" for user {} in context {} from queue", first.getTopic(), I(first.getUserId()), I(first.getContextId()));
+                do {
+                    polled.add(notification);
+                    LOGGER.debug("Polled notification \"{}\" for user {} in context {} from queue", notification.getTopic(), I(notification.getUserId()), I(notification.getContextId()));
+                    notification = submittedNotifcations.poll();
+                } while (null != notification);
+                onlyOne = false;
+            } else {
+                polled = Collections.singletonList(first);
+                LOGGER.debug("Polled notification \"{}\" for user {} in context {} from queue", first.getTopic(), I(first.getUserId()), I(first.getContextId()));
+                onlyOne = true;
+            }
 
-            if (scheduledNotifcations.isEmpty()) {
+            if (submittedNotifcations.isEmpty()) {
                 cancelTimerTask();
+                LOGGER.debug("Dropped polling timer task since queue is currently empty");
             }
         } finally {
             lock.unlock();
         }
 
-        if (1 == numAdded) {
+        if (onlyOne) {
             // There was only a single one available
             PushNotification first = polled.get(0);
-            int userId = first.getUserId();
-            int contextId = first.getContextId();
-            NotificationsHandler task = new NotificationsHandler(Collections.singleton(first), first.getTopic(), 1, userId, contextId);
-            if (false == tryExecuteTask(task, userId, contextId)) {
+            NotificationsHandler task = new NotificationsHandler(first);
+            if (tryExecuteTask(task, first.getUserId(), first.getContextId())) {
+                LOGGER.debug("Submitted delivery of polled notification with topic \"{}\" to user {} in context {}", first.getTopic(), I(first.getUserId()), I(first.getContextId()));
+            } else {
                 // Processor rejected task execution. Perform with current thread.
                 task.run();
             }
         } else {
             // Handle them all
-            Map<UserAndTopicKey, List<PushNotification>> polledNotifications = new LinkedHashMap<>();
+            Map<ContextAndTopicKey, Map<Integer, List<PushNotification>>> polledNotificationsPerContextAndTopic = new LinkedHashMap<>();
             for (PushNotification notification : polled) {
-                put(notification, polledNotifications);
+                put(notification, polledNotificationsPerContextAndTopic);
             }
-            for (Map.Entry<UserAndTopicKey, List<PushNotification>> entry : polledNotifications.entrySet()) {
-                UserAndTopicKey key = entry.getKey();
-                List<PushNotification> nots = entry.getValue();
-                NotificationsHandler task = new NotificationsHandler(nots, key.topic, nots.size(), key.userId, key.contextId);
-                if (false == tryExecuteTask(task, key.userId, key.contextId)) {
+            for (Map.Entry<ContextAndTopicKey, Map<Integer, List<PushNotification>>> entry : polledNotificationsPerContextAndTopic.entrySet()) {
+                ContextAndTopicKey key = entry.getKey();
+                Map<Integer, List<PushNotification>> notificationsPerUser = entry.getValue();
+                NotificationsHandler task = new NotificationsHandler(key.contextId, key.topic, notificationsPerUser);
+                if (tryExecuteTask(task, -1, key.contextId)) {
+                    if (LOGGER.isDebugEnabled()) {
+                        int[] userIds = Coll2i(notificationsPerUser.keySet());
+                        Object sUserIds = userIds.length <= 10 ? toStringObjectFor(userIds) : null;
+                        userIds = null;
+                        if (sUserIds == null) {
+                            LOGGER.debug("Submitted delivery of polled notifications with topic \"{}\" to users in context {}", key.topic, I(key.contextId));
+                        } else {
+                            LOGGER.debug("Submitted delivery of polled notifications with topic \"{}\" to users {} in context {}", key.topic, sUserIds, I(key.contextId));
+                        }
+                    }
+                } else {
                     // Processor rejected task execution. Perform with current thread.
                     task.run();
                 }
@@ -441,66 +475,105 @@ public class PushNotificationServiceImpl implements PushNotificationService {
         }
     }
 
-    private boolean tryExecuteTask(NotificationsHandler task, int userId, int contextId) {
+    private boolean tryExecuteTask(Runnable task, int userId, int contextId) {
         return processor.execute(UserAndContext.newInstance(userId, contextId), task);
     }
 
-    private void put(PushNotification notification, Map<UserAndTopicKey, List<PushNotification>> polledNotifications) {
-        UserAndTopicKey key = new UserAndTopicKey(notification);
-        List<PushNotification> l = polledNotifications.get(key);
-        if (null == l) {
-            l = new LinkedList<>();
-            polledNotifications.put(key, l);
+    private void put(PushNotification notification, Map<ContextAndTopicKey, Map<Integer, List<PushNotification>>> polledNotificationsPerUser) {
+        ContextAndTopicKey key = new ContextAndTopicKey(notification);
+        Map<Integer, List<PushNotification>> m = polledNotificationsPerUser.get(key);
+        if (null == m) {
+            m = new HashMap<Integer, List<PushNotification>>();
+            polledNotificationsPerUser.put(key, m);
         }
-        l.add(notification);
+        com.openexchange.tools.arrays.Collections.put(m, I(notification.getUserId()), notification);
     }
 
     /**
      * Handles given notifications.
      *
-     * @param notifications The notifications
-     * @param topic The topic (for all notifications)
-     * @param numOfNotifications The number of notifications provided by given iterator
-     * @param userId The user identifier (for all notifications)
      * @param contextId The context identifier (for all notifications)
+     * @param topic The topic (for all notifications)
+     * @param notificationsPerUser The notifications per user identifier
      * @throws OXException If handling fails
      */
-    protected void doHandle(Collection<PushNotification> notifications, String topic, int numOfNotifications, int userId, int contextId) throws OXException {
+    protected void doHandle(int contextId, String topic, Map<Integer, List<PushNotification>> notificationsPerUser) throws OXException {
+        // Determine available user identifiers from "notifications per user" map
+        int[] userIds = Coll2i(notificationsPerUser.keySet());
+        Object sUserIds = userIds.length <= 10 && LOGGER.isDebugEnabled() ? toStringObjectFor(userIds) : null;
+        if (sUserIds == null) {
+            LOGGER.debug("Delivering notifications for topic \"{}\" to users in context {}", topic, I(contextId));
+        } else {
+            LOGGER.debug("Delivering notifications for topic \"{}\" to users {} in context {}", topic, sUserIds, I(contextId));
+        }
+
         // Query appropriate hits
-        Hits hits = subscriptionRegistry.getInterestedSubscriptions(userId, contextId, topic);
+        Hits hits = subscriptionRegistry.getInterestedSubscriptions(userIds, contextId, topic);
         if (null == hits || hits.isEmpty()) {
-            LOGGER.debug("No subscriptions of interest for topic \"{}\" for user {} in context {}", topic, I(userId), I(contextId));
-            addNumOfProcessedNotifications(numOfNotifications);
+            if (sUserIds == null) {
+                LOGGER.debug("No subscriptions of interest for topic \"{}\" for users in context {}", topic, I(contextId));
+            } else {
+                LOGGER.debug("No subscriptions of interest for topic \"{}\" for users {} in context {}", topic, sUserIds, I(contextId));
+            }
+            addNumOfProcessedNotifications(getNumberOfNotificationsFrom(notificationsPerUser));
             return;
         }
 
-        // Transport each hit using associated transport
+        // Process by hits
+        if (sUserIds == null) {
+            LOGGER.debug("Determined subscriptions of interest for topic \"{}\" for users in context {}", topic, I(contextId));
+        } else {
+            LOGGER.debug("Determined subscriptions of interest for topic \"{}\" for users {} in context {}", topic, sUserIds, I(contextId));
+        }
+        for (Map.Entry<PushNotificationTransport, List<Hit>> entry : getHitsPerTransport(hits).entrySet()) {
+            // Get push matches for each notification
+            PushNotificationTransport transport = entry.getKey();
+            Map<PushNotification, List<PushMatch>> notifications = new HashMap<>();
+            for (Hit hit : entry.getValue()) {
+                List<PushMatch> matches = hit.getMatches();
+                for (PushMatch match : matches) {
+                    int userId = match.getUserId();
+                    if (isTransportAllowed(transport, topic, hit.getClient(), userId, contextId)) {
+                        List<PushNotification> notificationsForUser = notificationsPerUser.get(I(match.getUserId()));
+                        for (PushNotification notification : notificationsForUser) {
+                            if (matchesSourceToken(match, notification.getSourceToken())) {
+                                LOGGER.debug("Skipping push match for source token {}", notification.getSourceToken());
+                            } else {
+                                com.openexchange.tools.arrays.Collections.put(notifications, notification, match);
+                            }
+                        }
+                    } else {
+                        LOGGER.info("Transport '{}' not enabled for client '{}' to publish notification \"{}\" to user {} in context {}",
+                            transport.getId(), hit.getClient(), topic, I(userId), I(contextId));
+                    }
+                }
+            }
+
+            // Hand over to associated transport
+            if (sUserIds == null) {
+                LOGGER.debug("Going to use \"{}\" transport to publish notification(s) with topic \"{}\" to users in context {}", transport.getId(), topic, I(contextId));
+            } else {
+                LOGGER.debug("Going to use \"{}\" transport to publish notification(s) with topic \"{}\" to users {} in context {}", transport.getId(), topic, sUserIds, I(contextId));
+            }
+            transport.transport(notifications);
+        }
+
+        addNumOfProcessedNotifications(getNumberOfNotificationsFrom(notificationsPerUser));
+    }
+
+    private Map<PushNotificationTransport, List<Hit>> getHitsPerTransport(Hits hits) throws OXException {
+        Map<PushNotificationTransport, List<Hit>> hitsPerTransport = new HashMap<PushNotificationTransport, List<Hit>>();
         for (Hit hit : hits) {
             String client = hit.getClient();
             String transportId = hit.getTransportId();
             PushNotificationTransport transport = transportRegistry.getTransportFor(client, transportId);
             if (null == transport) {
-                LOGGER.info("No such transport '{}' for client '{}' to publish notification \"{}\" from user {} in context {}", transportId, client, topic, I(userId), I(contextId));
+                LOGGER.info("No transport '{}' for client '{}' available, skipping notificaton.", transportId, client);
             } else {
-                if (isTransportAllowed(transport, topic, client, userId, contextId)) {
-                    for (PushNotification notification : notifications) {
-                        List<PushMatch> matches = filterSourceToken(hit.getMatches(), notification.getSourceToken());
-                        if (null != matches && 0 < matches.size()) {
-                            LOGGER.debug("Trying to send notification \"{}\" via transport '{}' to client '{}' for user {} in context {}", topic, transportId, client, I(userId), I(contextId));
-                            try {
-                                transport.transport(notification, hit.getMatches());
-                            } catch (Exception e) {
-                                LOGGER.error("Failed to send notification \"{}\" via transport '{}' to client '{}' for user {} in context {}", topic, transportId, client, I(userId), I(contextId), e);
-                            }
-                        }
-                    }
-                } else {
-                    LOGGER.info("Transport '{}' not enabled for client '{}' to publish notification \"{}\" from user {} in context {}", transportId, client, topic, I(userId), I(contextId));
-                }
+                com.openexchange.tools.arrays.Collections.put(hitsPerTransport, transport, hit);
             }
         }
-
-        addNumOfProcessedNotifications(numOfNotifications);
+        return hitsPerTransport;
     }
 
     private boolean isTransportAllowed(PushNotificationTransport transport, String topic, String client, int userId, int contextId) {
@@ -512,16 +585,16 @@ public class PushNotificationServiceImpl implements PushNotificationService {
         }
     }
 
-    private List<PushMatch> filterSourceToken(List<PushMatch> matches, String sourceToken) {
-        if (null != sourceToken && null != matches) {
-            for (Iterator<PushMatch> iterator = matches.iterator(); iterator.hasNext();) {
-                if (sourceToken.equals(iterator.next().getToken())) {
-                    LOGGER.debug("Skipping push match for source token {}", sourceToken);
-                    iterator.remove();
-                }
-            }
+    private boolean matchesSourceToken(PushMatch match, String sourceToken) {
+        return null != sourceToken && null != match && sourceToken.equals(match.getToken());
+    }
+
+    private int getNumberOfNotificationsFrom(Map<Integer, List<PushNotification>> notificationsPerUser) {
+        int count = 0;
+        for (List<PushNotification> notifications : notificationsPerUser.values()) {
+            count += notifications.size();
         }
-        return matches;
+        return count;
     }
 
     private void addNumOfProcessedNotifications(int numOfNotifications) {
@@ -532,31 +605,33 @@ public class PushNotificationServiceImpl implements PushNotificationService {
 
     private final class NotificationsHandler implements Runnable {
 
-        private final Collection<PushNotification> notifications;
+        private final Map<Integer, List<PushNotification>> notificationsPerUser;
         private final String topic;
-        private final int userId;
         private final int contextId;
-        private final int numOfNotifications;
 
         /**
          * Initializes a new {@link PushNotificationServiceImpl.NotificationsHandler}.
          */
-        NotificationsHandler(Collection<PushNotification> notifications, String topic, int numOfNotifications, int userId, int contextId) {
-            super();
-            this.notifications = notifications;
-            this.topic = topic;
-            this.numOfNotifications = numOfNotifications;
-            this.userId = userId;
-            this.contextId = contextId;
+        NotificationsHandler(PushNotification notification) {
+            this(notification.getContextId(), notification.getTopic(), Collections.singletonMap(I(notification.getUserId()), Collections.singletonList(notification)));
+        }
 
+        /**
+         * Initializes a new {@link PushNotificationServiceImpl.NotificationsHandler}.
+         */
+        NotificationsHandler(int contextId, String topic, Map<Integer, List<PushNotification>> notificationsPerUser) {
+            super();
+            this.notificationsPerUser = notificationsPerUser;
+            this.topic = topic;
+            this.contextId = contextId;
         }
 
         @Override
         public void run() {
             try {
-                doHandle(notifications, topic, numOfNotifications, userId, contextId);
+                doHandle(contextId, topic, notificationsPerUser);
             } catch (Exception e) {
-                LOGGER.error("Failed to handle notification(s) with topic {} for user {} in context {}", topic, I(userId), I(contextId), e);
+                LOGGER.error("Failed to handle notification(s) with topic \"{}\" in context {}", topic, I(contextId), e);
             }
         }
     }
