@@ -79,11 +79,11 @@ public class PnsCreateTableTask extends UpdateTaskAdapter {
     @Override
     public void perform(PerformParameters params) throws OXException {
         Connection con = params.getConnection();
-        boolean rollback = false;
+        int rollback = 0;
         PreparedStatement stmt = null;
         try {
             con.setAutoCommit(false);
-            rollback = true;
+            rollback = 1;
 
             if (!tableExists(con, "pns_subscription")) {
                 stmt = con.prepareStatement(CreatePnsSubscriptionTable.getTableSubscription());
@@ -105,15 +105,17 @@ public class PnsCreateTableTask extends UpdateTaskAdapter {
             }
 
             con.commit();
-            rollback = false;
+            rollback = 2;
         } catch (SQLException e) {
             throw UpdateExceptionCodes.SQL_PROBLEM.create(e, e.getMessage());
         } finally {
             closeSQLStuff(stmt);
-            if (rollback) {
-                Databases.rollback(con);
+            if (rollback > 0) {
+                if (rollback == 1) {
+                    Databases.rollback(con);
+                }
+                Databases.autocommit(con);
             }
-            Databases.autocommit(con);
         }
     }
 }
